@@ -3,6 +3,7 @@
 import asyncio
 import base64
 import hashlib
+import json
 import os
 import sys
 import traceback
@@ -11,20 +12,26 @@ import warnings
 from multidict import CIMultiDict, MultiDict, MultiDictProxy, istr
 from yarl import URL
 
-import aiohttp
-
-from . import hdrs, helpers
-from ._ws_impl import WS_KEY, WebSocketParser, WebSocketWriter
+from . import connector as connector_mod
+from . import client_exceptions, client_reqrep, hdrs, http, payload
+from .client_exceptions import *  # noqa
+from .client_exceptions import (ClientError, ClientOSError,
+                                ClientResponseError, ServerTimeoutError,
+                                WSServerHandshakeError)
+from .client_reqrep import *  # noqa
 from .client_reqrep import ClientRequest, ClientResponse
 from .client_ws import ClientWebSocketResponse
+from .connector import *  # noqa
+from .connector import TCPConnector
 from .cookiejar import CookieJar
-from .errors import WSServerHandshakeError
-from .helpers import TimeoutHandle, TimerContext, TimeService
+from .helpers import PY_35, CeilTimeout, TimeoutHandle, deprecated_noop
+from .http import WS_KEY, WebSocketReader, WebSocketWriter
+from .streams import FlowControlDataQueue
 
-__all__ = ('ClientSession', 'request', 'get', 'options', 'head',
-           'delete', 'post', 'put', 'patch', 'ws_connect')
-
-PY_35 = sys.version_info >= (3, 5)
+__all__ = (client_exceptions.__all__ +  # noqa
+           client_reqrep.__all__ +  # noqa
+           connector_mod.__all__ +  # noqa
+           ('ClientSession', 'ClientWebSocketResponse', 'request'))
 
 
 # 5 Minute default read and connect timeout
@@ -39,11 +46,12 @@ class ClientSession:
 
     def __init__(self, *, connector=None, loop=None, cookies=None,
                  headers=None, skip_auto_headers=None,
-                 auth=None, request_class=ClientRequest,
-                 response_class=ClientResponse,
+                 auth=None, json_serialize=json.dumps,
+                 request_class=ClientRequest, response_class=ClientResponse,
                  ws_response_class=ClientWebSocketResponse,
-                 version=aiohttp.HttpVersion11,
-                 cookie_jar=None, read_timeout=None, time_service=None):
+                 version=http.HttpVersion11,
+                 cookie_jar=None, connector_owner=True, raise_for_status=False,
+                 read_timeout=None, conn_timeout=None):
 
         implicit_loop = False
         if loop is None:
@@ -54,7 +62,7 @@ class ClientSession:
                 loop = asyncio.get_event_loop()
 
         if connector is None:
-            connector = aiohttp.TCPConnector(loop=loop)
+            connector = TCPConnector(loop=loop)
 
         if connector._loop is not loop:
             raise RuntimeError(
@@ -83,9 +91,13 @@ class ClientSession:
         if cookies is not None:
             self._cookie_jar.update_cookies(cookies)
         self._connector = connector
+        self._connector_owner = connector_owner
         self._default_auth = auth
         self._version = version
+        self._json_serialize = json_serialize
         self._read_timeout = read_timeout
+        self._conn_timeout = conn_timeout
+        self._raise_for_status = raise_for_status
 
         # Convert to list of tuples
         if headers:
@@ -103,13 +115,6 @@ class ClientSession:
         self._response_class = response_class
         self._ws_response_class = ws_response_class
 
-        if time_service is not None:
-            self._time_service_owner = False
-            self._time_service = time_service
-        else:
-            self._time_service_owner = True
-            self._time_service = TimeService(self._loop)
-
     def __del__(self, _warnings=warnings):
         if not self.closed:
             self.close()
@@ -122,10 +127,6 @@ class ClientSession:
                 context['source_traceback'] = self._source_traceback
             self._loop.call_exception_handler(context)
 
-    @property
-    def time_service(self):
-        return self._time_service
-
     def request(self, method, url, **kwargs):
         """Perform HTTP request."""
         return _RequestContextManager(self._request(method, url, **kwargs))
@@ -134,13 +135,13 @@ class ClientSession:
     def _request(self, method, url, *,
                  params=None,
                  data=None,
+                 json=None,
                  headers=None,
                  skip_auto_headers=None,
                  auth=None,
                  allow_redirects=True,
                  max_redirects=10,
-                 encoding='utf-8',
-                 version=None,
+                 encoding=None,
                  compress=None,
                  chunked=None,
                  expect100=False,
@@ -153,17 +154,28 @@ class ClientSession:
         # set the default to None because we need to detect if the user wants
         # to use the existing timeouts by setting timeout to None.
 
-        if version is not None:
-            warnings.warn("HTTP version should be specified "
-                          "by ClientSession constructor", DeprecationWarning)
-        else:
-            version = self._version
+        if encoding is not None:
+            warnings.warn(
+                "encoding parameter is not supported, "
+                "please use FormData(charset='utf-8') instead",
+                DeprecationWarning)
 
         if self.closed:
             raise RuntimeError('Session is closed')
 
+        if data is not None and json is not None:
+            raise ValueError(
+                'data and json parameters can not be used at the same time')
+        elif json is not None:
+            data = payload.JsonPayload(json, dumps=self._json_serialize)
+
+        if not isinstance(chunked, bool) and chunked is not None:
+            warnings.warn(
+                'Chunk size is deprecated #1615', DeprecationWarning)
+
         redirects = 0
         history = []
+        version = self._version
 
         # Merge with default headers and transform to CIMultiDict
         headers = self._prepare_headers(headers)
@@ -185,104 +197,125 @@ class ClientSession:
         if proxy is not None:
             proxy = URL(proxy)
 
-        # request timeout
-        if timeout is None:
-            timeout = self._read_timeout
-        if timeout is None:
-            timeout = self._connector.conn_timeout
-        elif self._connector.conn_timeout is not None:
-            timeout = max(timeout, self._connector.conn_timeout)
-
         # timeout is cumulative for all request operations
         # (request, redirects, responses, data consuming)
-        tm = TimeoutHandle(timeout)
-        handle = tm.handle(self._loop)
+        tm = TimeoutHandle(
+            self._loop, timeout if timeout is not None else self._read_timeout)
+        handle = tm.start()
 
-        timer = TimerContext(self._loop, tm)
-        with timer:
-            while True:
-                url = URL(url).with_fragment(None)
+        timer = tm.timer()
+        try:
+            with timer:
+                while True:
+                    url = URL(url).with_fragment(None)
+                    cookies = self._cookie_jar.filter_cookies(url)
 
-                cookies = self._cookie_jar.filter_cookies(url)
+                    req = self._request_class(
+                        method, url, params=params, headers=headers,
+                        skip_auto_headers=skip_headers, data=data,
+                        cookies=cookies, auth=auth, version=version,
+                        compress=compress, chunked=chunked,
+                        expect100=expect100, loop=self._loop,
+                        response_class=self._response_class,
+                        proxy=proxy, proxy_auth=proxy_auth, timer=timer)
 
-                req = self._request_class(
-                    method, url, params=params, headers=headers,
-                    skip_auto_headers=skip_headers, data=data,
-                    cookies=cookies, encoding=encoding,
-                    auth=auth, version=version, compress=compress,
-                    chunked=chunked, expect100=expect100,
-                    loop=self._loop, response_class=self._response_class,
-                    proxy=proxy, proxy_auth=proxy_auth, timer=timer)
-
-                conn = yield from self._connector.connect(req)
-                conn.writer.set_tcp_nodelay(True)
-                try:
-                    resp = req.send(conn.writer, conn.reader)
+                    # connection timeout
                     try:
-                        yield from resp.start(conn, read_until_eof)
-                    except:
-                        resp.close()
-                        conn.close()
+                        with CeilTimeout(self._conn_timeout, loop=self._loop):
+                            conn = yield from self._connector.connect(req)
+                    except asyncio.TimeoutError as exc:
+                        raise ServerTimeoutError(
+                            'Connection timeout '
+                            'to host {0}'.format(url)) from exc
+
+                    conn.writer.set_tcp_nodelay(True)
+                    try:
+                        resp = req.send(conn)
+                        try:
+                            yield from resp.start(conn, read_until_eof)
+                        except:
+                            resp.close()
+                            conn.close()
+                            raise
+                    except ClientError:
                         raise
-                except (aiohttp.HttpProcessingError,
-                        aiohttp.ServerDisconnectedError) as exc:
-                    raise aiohttp.ClientResponseError() from exc
-                except OSError as exc:
-                    raise aiohttp.ClientOSError(*exc.args) from exc
+                    except http.HttpProcessingError as exc:
+                        raise ClientResponseError(
+                            code=exc.code,
+                            message=exc.message, headers=exc.headers) from exc
+                    except OSError as exc:
+                        raise ClientOSError(*exc.args) from exc
 
-                self._cookie_jar.update_cookies(resp.cookies, resp.url_obj)
+                    self._cookie_jar.update_cookies(resp.cookies, resp.url)
 
-                # redirects
-                if resp.status in (301, 302, 303, 307) and allow_redirects:
-                    redirects += 1
-                    history.append(resp)
-                    if max_redirects and redirects >= max_redirects:
-                        resp.close()
-                        break
-                    else:
-                        yield from resp.release()
+                    # redirects
+                    if resp.status in (301, 302, 303, 307) and allow_redirects:
+                        redirects += 1
+                        history.append(resp)
+                        if max_redirects and redirects >= max_redirects:
+                            resp.close()
+                            break
+                        else:
+                            resp.release()
 
-                    # For 301 and 302, mimic IE behaviour, now changed in RFC.
-                    # Info: https://github.com/kennethreitz/requests/pull/269
-                    if (resp.status == 303 and resp.method != hdrs.METH_HEAD) \
-                       or (resp.status in (301, 302) and
-                           resp.method == hdrs.METH_POST):
-                        method = hdrs.METH_GET
-                        data = None
-                        if headers.get(hdrs.CONTENT_LENGTH):
-                            headers.pop(hdrs.CONTENT_LENGTH)
+                        # For 301 and 302, mimic IE, now changed in RFC
+                        # https://github.com/kennethreitz/requests/pull/269
+                        if (resp.status == 303 and
+                                resp.method != hdrs.METH_HEAD) \
+                                or (resp.status in (301, 302) and
+                                    resp.method == hdrs.METH_POST):
+                            method = hdrs.METH_GET
+                            data = None
+                            if headers.get(hdrs.CONTENT_LENGTH):
+                                headers.pop(hdrs.CONTENT_LENGTH)
 
-                    r_url = (resp.headers.get(hdrs.LOCATION) or
-                             resp.headers.get(hdrs.URI))
-                    if r_url is None:
-                        raise RuntimeError("{0.method} {0.url_obj} returns "
-                                           "a redirect [{0.status}] status "
-                                           "but response lacks a Location "
-                                           "or URI HTTP header".format(resp))
-                    r_url = URL(r_url)
+                        r_url = (resp.headers.get(hdrs.LOCATION) or
+                                 resp.headers.get(hdrs.URI))
+                        if r_url is None:
+                            raise RuntimeError(
+                                "{0.method} {0.url} returns "
+                                "a redirect [{0.status}] status "
+                                "but response lacks a Location "
+                                "or URI HTTP header".format(resp))
+                        r_url = URL(r_url)
 
-                    scheme = r_url.scheme
-                    if scheme not in ('http', 'https', ''):
-                        resp.close()
-                        raise ValueError('Can redirect only to http or https')
-                    elif not scheme:
-                        r_url = url.join(r_url)
+                        scheme = r_url.scheme
+                        if scheme not in ('http', 'https', ''):
+                            resp.close()
+                            raise ValueError(
+                                'Can redirect only to http or https')
+                        elif not scheme:
+                            r_url = url.join(r_url)
 
-                    url = r_url
-                    params = None
-                    yield from resp.release()
-                    continue
+                        url = r_url
+                        params = None
+                        resp.release()
+                        continue
 
-                break
+                    break
 
-        if handle is not None:
-            if resp.connection is not None:
-                resp.connection.add_callback(handle.cancel)
-            else:
+            # check response status
+            if self._raise_for_status:
+                resp.raise_for_status()
+
+            # register connection
+            if handle is not None:
+                if resp.connection is not None:
+                    resp.connection.add_callback(handle.cancel)
+                else:
+                    handle.cancel()
+
+            resp._history = tuple(history)
+            return resp
+
+        except:
+            # cleanup timer
+            tm.close()
+            if handle:
                 handle.cancel()
+                handle = None
 
-        resp._history = tuple(history)
-        return resp
+            raise
 
     def ws_connect(self, url, *,
                    protocols=(),
@@ -395,7 +428,10 @@ class ClientSession:
                         protocol = proto
                         break
 
-            reader = resp.connection.reader.set_parser(WebSocketParser)
+            proto = resp.connection.protocol
+            reader = FlowControlDataQueue(
+                proto, limit=2 ** 16, loop=self._loop)
+            proto.set_parser(WebSocketReader(reader), reader)
             resp.connection.writer.set_tcp_nodelay(True)
             writer = WebSocketWriter(resp.connection.writer, use_mask=True)
         except Exception:
@@ -410,7 +446,6 @@ class ClientSession:
                                            autoclose,
                                            autoping,
                                            self._loop,
-                                           time_service=self.time_service,
                                            receive_timeout=receive_timeout,
                                            heartbeat=heartbeat)
 
@@ -485,15 +520,11 @@ class ClientSession:
         Release all acquired resources.
         """
         if not self.closed:
-            self._connector.close()
+            if self._connector_owner:
+                self._connector.close()
             self._connector = None
 
-            if self._time_service_owner:
-                self._time_service.close()
-
-        ret = helpers.create_future(self._loop)
-        ret.set_result(None)
-        return ret
+        return deprecated_noop('ClientSession.close() is not coroutine')
 
     @property
     def closed(self):
@@ -544,7 +575,7 @@ class ClientSession:
 
         @asyncio.coroutine
         def __aexit__(self, exc_type, exc_val, exc_tb):
-            yield from self.close()
+            self.close()
 
 
 if PY_35:
@@ -613,7 +644,7 @@ class _RequestContextManager(_BaseRequestContextManager):
             # would like to close a connection you must do that
             # explicitly.  Otherwise connection error handling should kick in
             # and close/recycle the connection as required.
-            yield from self._resp.release()
+            self._resp.release()
 
 
 class _WSRequestContextManager(_BaseRequestContextManager):
@@ -636,7 +667,7 @@ class _DetachedRequestContextManager(_RequestContextManager):
         try:
             return (yield from self._coro)
         except:
-            yield from self._session.close()
+            self._session.close()
             raise
 
     if PY_35:
@@ -644,54 +675,41 @@ class _DetachedRequestContextManager(_RequestContextManager):
             try:
                 return (yield from self._coro)
             except:
-                yield from self._session.close()
+                self._session.close()
                 raise
 
     def __del__(self):
         self._session.detach()
 
 
-class _DetachedWSRequestContextManager(_WSRequestContextManager):
-
-    __slots__ = _WSRequestContextManager.__slots__ + ('_session', )
-
-    def __init__(self, coro, session):
-        super().__init__(coro)
-        self._session = session
-
-    def __del__(self):
-        self._session.detach()
-
-
 def request(method, url, *,
             params=None,
             data=None,
+            json=None,
             headers=None,
             skip_auto_headers=None,
             cookies=None,
             auth=None,
             allow_redirects=True,
             max_redirects=10,
-            encoding='utf-8',
-            version=None,
+            encoding=None,
+            version=http.HttpVersion11,
             compress=None,
             chunked=None,
             expect100=False,
             connector=None,
             loop=None,
             read_until_eof=True,
-            request_class=None,
-            response_class=None,
             proxy=None,
             proxy_auth=None):
     """Constructs and sends a request. Returns response object.
-
     method - HTTP method
     url - request url
     params - (optional) Dictionary or bytes to be sent in the query
       string of the new request
     data - (optional) Dictionary, bytes, or file-like object to
       send in the body of the request
+    json - (optional) Any json compatibile python object
     headers - (optional) Dictionary of HTTP Headers to send with
       the request
     cookies - (optional) Dict object to send with the request
@@ -708,45 +726,34 @@ def request(method, url, *,
        connection pooling.
     read_until_eof - Read response until eof if response
        does not have Content-Length header.
-    request_class - (optional) Custom Request class implementation.
-    response_class - (optional) Custom Response class implementation.
     loop - Optional event loop.
-
     Usage::
-
       >>> import aiohttp
       >>> resp = yield from aiohttp.request('GET', 'http://python.org/')
       >>> resp
       <ClientResponse(python.org/) [200]>
       >>> data = yield from resp.read()
-
     """
+    connector_owner = False
     if connector is None:
-        connector = aiohttp.TCPConnector(loop=loop, force_close=True)
+        connector_owner = True
+        connector = TCPConnector(loop=loop, force_close=True)
 
-    kwargs = {}
+    session = ClientSession(
+        loop=loop, cookies=cookies, version=version,
+        connector=connector, connector_owner=connector_owner)
 
-    if request_class is not None:
-        kwargs['request_class'] = request_class
-
-    if response_class is not None:
-        kwargs['response_class'] = response_class
-
-    session = ClientSession(loop=loop,
-                            cookies=cookies,
-                            connector=connector,
-                            **kwargs)
     return _DetachedRequestContextManager(
         session._request(method, url,
                          params=params,
                          data=data,
+                         json=json,
                          headers=headers,
                          skip_auto_headers=skip_auto_headers,
                          auth=auth,
                          allow_redirects=allow_redirects,
                          max_redirects=max_redirects,
                          encoding=encoding,
-                         version=version,
                          compress=compress,
                          chunked=chunked,
                          expect100=expect100,
@@ -754,64 +761,3 @@ def request(method, url, *,
                          proxy=proxy,
                          proxy_auth=proxy_auth,),
         session=session)
-
-
-def get(url, **kwargs):
-    warnings.warn("Use ClientSession().get() instead", DeprecationWarning)
-    return request(hdrs.METH_GET, url, **kwargs)
-
-
-def options(url, **kwargs):
-    warnings.warn("Use ClientSession().options() instead", DeprecationWarning)
-    return request(hdrs.METH_OPTIONS, url, **kwargs)
-
-
-def head(url, **kwargs):
-    warnings.warn("Use ClientSession().head() instead", DeprecationWarning)
-    return request(hdrs.METH_HEAD, url, **kwargs)
-
-
-def post(url, **kwargs):
-    warnings.warn("Use ClientSession().post() instead", DeprecationWarning)
-    return request(hdrs.METH_POST, url, **kwargs)
-
-
-def put(url, **kwargs):
-    warnings.warn("Use ClientSession().put() instead", DeprecationWarning)
-    return request(hdrs.METH_PUT, url, **kwargs)
-
-
-def patch(url, **kwargs):
-    warnings.warn("Use ClientSession().patch() instead", DeprecationWarning)
-    return request(hdrs.METH_PATCH, url, **kwargs)
-
-
-def delete(url, **kwargs):
-    warnings.warn("Use ClientSession().delete() instead", DeprecationWarning)
-    return request(hdrs.METH_DELETE, url, **kwargs)
-
-
-def ws_connect(url, *, protocols=(), timeout=10.0, connector=None, auth=None,
-               ws_response_class=ClientWebSocketResponse, autoclose=True,
-               autoping=True, loop=None, origin=None, headers=None):
-
-    warnings.warn("Use ClientSession().ws_connect() instead",
-                  DeprecationWarning)
-    if loop is None:
-        loop = asyncio.get_event_loop()
-
-    if connector is None:
-        connector = aiohttp.TCPConnector(loop=loop, force_close=True)
-
-    session = aiohttp.ClientSession(loop=loop, connector=connector, auth=auth,
-                                    ws_response_class=ws_response_class,
-                                    headers=headers)
-
-    return _DetachedWSRequestContextManager(
-        session._ws_connect(url,
-                            protocols=protocols,
-                            timeout=timeout,
-                            autoclose=autoclose,
-                            autoping=autoping,
-                            origin=origin),
-        session=session)
