@@ -6,21 +6,20 @@ import binascii
 import cgi
 import datetime
 import functools
-import heapq
-import io
 import os
 import re
 import sys
 import time
 import warnings
+import weakref
 from collections import MutableSequence, namedtuple
 from functools import total_ordering
+from math import ceil
 from pathlib import Path
 from time import gmtime
-from urllib.parse import urlencode
+from urllib.parse import quote
 
 from async_timeout import timeout
-from multidict import MultiDict, MultiDictProxy
 
 from . import hdrs
 
@@ -29,6 +28,9 @@ try:
 except ImportError:
     ensure_future = asyncio.async
 
+PY_34 = sys.version_info < (3, 5)
+PY_35 = sys.version_info >= (3, 5)
+PY_352 = sys.version_info >= (3, 5, 2)
 
 if sys.version_info >= (3, 4, 3):
     from http.cookies import SimpleCookie  # noqa
@@ -36,12 +38,48 @@ else:
     from .backport_cookies import SimpleCookie  # noqa
 
 
-__all__ = ('BasicAuth', 'create_future', 'FormData', 'parse_mimetype',
-           'Timeout', 'ensure_future')
+__all__ = ('BasicAuth', 'create_future', 'parse_mimetype',
+           'Timeout', 'ensure_future', 'noop')
 
 
 sentinel = object()
 Timeout = timeout
+NO_EXTENSIONS = bool(os.environ.get('AIOHTTP_NO_EXTENSIONS'))
+
+CHAR = set(chr(i) for i in range(0, 128))
+CTL = set(chr(i) for i in range(0, 32)) | {chr(127), }
+SEPARATORS = {'(', ')', '<', '>', '@', ',', ';', ':', '\\', '"', '/', '[', ']',
+              '?', '=', '{', '}', ' ', chr(9)}
+TOKEN = CHAR ^ CTL ^ SEPARATORS
+
+
+if sys.version_info < (3, 5):
+    noop = tuple
+
+    coroutines = asyncio.coroutines
+    old_debug = coroutines._DEBUG
+    coroutines._DEBUG = False
+
+    @asyncio.coroutine
+    def deprecated_noop(message):
+        warnings.warn(message, DeprecationWarning, stacklevel=3)
+
+    coroutines._DEBUG = old_debug
+
+else:
+    coroutines = asyncio.coroutines
+    old_debug = coroutines._DEBUG
+    coroutines._DEBUG = False
+
+    @asyncio.coroutine
+    def noop(*args, **kwargs):
+        return
+
+    @asyncio.coroutine
+    def deprecated_noop(message):
+        warnings.warn(message, DeprecationWarning, stacklevel=3)
+
+    coroutines._DEBUG = old_debug
 
 
 class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
@@ -92,130 +130,16 @@ class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
         return 'Basic %s' % base64.b64encode(creds).decode(self.encoding)
 
 
-def create_future(loop):
-    """Compatibility wrapper for the loop.create_future() call introduced in
-    3.5.2."""
-    if hasattr(loop, 'create_future'):
+if PY_352:
+    def create_future(loop):
         return loop.create_future()
-    else:
+else:
+    def create_future(loop):
+        """Compatibility wrapper for the loop.create_future() call introduced in
+        3.5.2."""
         return asyncio.Future(loop=loop)
 
 
-class FormData:
-    """Helper class for multipart/form-data and
-    application/x-www-form-urlencoded body generation."""
-
-    def __init__(self, fields=(), quote_fields=True):
-        from . import multipart
-        self._writer = multipart.MultipartWriter('form-data')
-        self._fields = []
-        self._is_multipart = False
-        self._quote_fields = quote_fields
-
-        if isinstance(fields, dict):
-            fields = list(fields.items())
-        elif not isinstance(fields, (list, tuple)):
-            fields = (fields,)
-        self.add_fields(*fields)
-
-    @property
-    def is_multipart(self):
-        return self._is_multipart
-
-    @property
-    def content_type(self):
-        if self._is_multipart:
-            return self._writer.headers[hdrs.CONTENT_TYPE]
-        else:
-            return 'application/x-www-form-urlencoded'
-
-    def add_field(self, name, value, *, content_type=None, filename=None,
-                  content_transfer_encoding=None):
-
-        if isinstance(value, io.IOBase):
-            self._is_multipart = True
-        elif isinstance(value, (bytes, bytearray, memoryview)):
-            if filename is None and content_transfer_encoding is None:
-                filename = name
-
-        type_options = MultiDict({'name': name})
-        if filename is not None and not isinstance(filename, str):
-            raise TypeError('filename must be an instance of str. '
-                            'Got: %s' % filename)
-        if filename is None and isinstance(value, io.IOBase):
-            filename = guess_filename(value, name)
-        if filename is not None:
-            type_options['filename'] = filename
-            self._is_multipart = True
-
-        headers = {}
-        if content_type is not None:
-            if not isinstance(content_type, str):
-                raise TypeError('content_type must be an instance of str. '
-                                'Got: %s' % content_type)
-            headers[hdrs.CONTENT_TYPE] = content_type
-            self._is_multipart = True
-        if content_transfer_encoding is not None:
-            if not isinstance(content_transfer_encoding, str):
-                raise TypeError('content_transfer_encoding must be an instance'
-                                ' of str. Got: %s' % content_transfer_encoding)
-            headers[hdrs.CONTENT_TRANSFER_ENCODING] = content_transfer_encoding
-            self._is_multipart = True
-
-        self._fields.append((type_options, headers, value))
-
-    def add_fields(self, *fields):
-        to_add = list(fields)
-
-        while to_add:
-            rec = to_add.pop(0)
-
-            if isinstance(rec, io.IOBase):
-                k = guess_filename(rec, 'unknown')
-                self.add_field(k, rec)
-
-            elif isinstance(rec, (MultiDictProxy, MultiDict)):
-                to_add.extend(rec.items())
-
-            elif isinstance(rec, (list, tuple)) and len(rec) == 2:
-                k, fp = rec
-                self.add_field(k, fp)
-
-            else:
-                raise TypeError('Only io.IOBase, multidict and (name, file) '
-                                'pairs allowed, use .add_field() for passing '
-                                'more complex parameters, got {!r}'
-                                .format(rec))
-
-    def _gen_form_urlencoded(self, encoding):
-        # form data (x-www-form-urlencoded)
-        data = []
-        for type_options, _, value in self._fields:
-            data.append((type_options['name'], value))
-
-        data = urlencode(data, doseq=True)
-        return data.encode(encoding)
-
-    def _gen_form_data(self, *args, **kwargs):
-        """Encode a list of fields using the multipart/form-data MIME format"""
-        for dispparams, headers, value in self._fields:
-            part = self._writer.append(value, headers)
-            if dispparams:
-                part.set_content_disposition(
-                    'form-data', quote_fields=self._quote_fields, **dispparams
-                )
-                # FIXME cgi.FieldStorage doesn't likes body parts with
-                # Content-Length which were sent via chunked transfer encoding
-                part.headers.pop(hdrs.CONTENT_LENGTH, None)
-        yield from self._writer.serialize()
-
-    def __call__(self, encoding):
-        if self._is_multipart:
-            return self._gen_form_data(encoding)
-        else:
-            return self._gen_form_urlencoded(encoding)
-
-
 def parse_mimetype(mimetype):
     """Parses a MIME type into its components.
 
@@ -260,6 +184,33 @@ def guess_filename(obj, default=None):
     return default
 
 
+def content_disposition_header(disptype, quote_fields=True, **params):
+    """Sets ``Content-Disposition`` header.
+
+    :param str disptype: Disposition type: inline, attachment, form-data.
+                         Should be valid extension token (see RFC 2183)
+    :param dict params: Disposition params
+    """
+    if not disptype or not (TOKEN > set(disptype)):
+        raise ValueError('bad content disposition type {!r}'
+                         ''.format(disptype))
+
+    value = disptype
+    if params:
+        lparams = []
+        for key, val in params.items():
+            if not key or not (TOKEN > set(key)):
+                raise ValueError('bad content disposition parameter'
+                                 ' {!r}={!r}'.format(key, val))
+            qval = quote(val, '') if quote_fields else val
+            lparams.append((key, '"%s"' % qval))
+            if key == 'filename':
+                lparams.append(('filename*', "utf-8''" + qval))
+        sparams = '; '.join('='.join(pair) for pair in lparams)
+        value = '; '.join((value, sparams))
+    return value
+
+
 class AccessLogger:
     """Helper object to log access.
 
@@ -423,7 +374,7 @@ class AccessLogger:
 
     @staticmethod
     def _format_O(args):
-        return args[2].output_length
+        return args[2].body_length
 
     @staticmethod
     def _format_T(args):
@@ -486,14 +437,17 @@ class reify:
         self.name = wrapped.__name__
 
     def __get__(self, inst, owner, _sentinel=sentinel):
-        if inst is None:
-            return self
-        val = inst._cache.get(self.name, _sentinel)
-        if val is not _sentinel:
-            return val
-        val = self.wrapped(inst)
-        inst._cache[self.name] = val
-        return val
+        try:
+            try:
+                return inst._cache[self.name]
+            except KeyError:
+                val = self.wrapped(inst)
+                inst._cache[self.name] = val
+                return val
+        except AttributeError:
+            if inst is None:
+                return self
+            raise
 
     def __set__(self, inst, value):
         raise AttributeError("reified property is read-only")
@@ -534,17 +488,6 @@ def is_ip_address(host):
                         .format(host, type(host)))
 
 
-def _get_kwarg(kwargs, old, new, value):
-    val = kwargs.pop(old, sentinel)
-    if val is not sentinel:
-        warnings.warn("{} is deprecated, use {} instead".format(old, new),
-                      DeprecationWarning,
-                      stacklevel=3)
-        return val
-    else:
-        return value
-
-
 @total_ordering
 class FrozenList(MutableSequence):
 
@@ -560,6 +503,7 @@ class FrozenList(MutableSequence):
 
     def freeze(self):
         self._frozen = True
+        self._items = tuple(self._items)
 
     def __getitem__(self, index):
         return self._items[index]
@@ -595,12 +539,6 @@ class FrozenList(MutableSequence):
         self._items.insert(pos, item)
 
 
-class TimerHandle(asyncio.TimerHandle):
-
-    def cancel(self):
-        asyncio.Handle.cancel(self)
-
-
 class TimeService:
 
     def __init__(self, loop, *, interval=1.0):
@@ -611,23 +549,15 @@ class TimeService:
         self._count = 0
         self._strtime = None
         self._cb = loop.call_at(self._loop_time + self._interval, self._on_cb)
-        self._scheduled = []
 
     def close(self):
         if self._cb:
             self._cb.cancel()
 
-        # cancel all scheduled handles
-        for handle in self._scheduled:
-            handle.cancel()
-
         self._cb = None
-        self._scheduled = []
         self._loop = None
 
     def _on_cb(self, reset_count=10*60):
-        self._loop_time = self._loop.time()
-
         if self._count >= reset_count:
             # reset timer every 10 minutes
             self._count = 0
@@ -635,21 +565,8 @@ class TimeService:
         else:
             self._time += self._interval
 
-        # Handle 'later' callbacks that are ready.
-        ready = []
-        end_time = self._loop_time
-        while self._scheduled:
-            handle = self._scheduled[0]
-            if handle._when >= end_time:
-                break
-            handle = heapq.heappop(self._scheduled)
-            ready.append(handle)
-
-        for handle in ready:
-            if not handle._cancelled:
-                handle._run()
-
         self._strtime = None
+        self._loop_time = ceil(self._loop.time())
         self._cb = self._loop.call_at(
             self._loop_time + self._interval, self._on_cb)
 
@@ -676,50 +593,67 @@ class TimeService:
             self._strtime = s = self._format_date_time()
         return self._strtime
 
+    @property
     def loop_time(self):
         return self._loop_time
 
-    def call_later(self, delay, callback, *args):
-        """Arrange for a callback to be called at a given time.
 
-        Return a Handle: an opaque object with a cancel() method that
-        can be used to cancel the call.
+def _weakref_handle(ref):
+    cb = ref()
+    if cb is not None:
+        cb()
 
-        The delay can be an int or float, expressed in seconds.  It is
-        always relative to the current time.
 
-        Any positional arguments after the callback will be passed to
-        the callback when it is called.
+def weakref_handle(cb, timeout, loop, ceil_timeout=True):
+    if timeout is not None and timeout > 0:
+        when = loop.time() + timeout
+        if ceil_timeout:
+            when = ceil(when)
 
-        Time resolution is aproximatly one second.
-        """
-        return self._call_at(self._loop_time + delay, callback, *args)
+        return loop.call_at(when, _weakref_handle, weakref.ref(cb))
 
-    def _call_at(self, when, callback, *args):
-        """Like call_later(), but uses an absolute time.
 
-        Absolute time corresponds to the loop's time() method.
-        """
-        timer = TimerHandle(when, callback, args, self._loop)
-        heapq.heappush(self._scheduled, timer)
+def call_later(cb, timeout, loop):
+    if timeout is not None and timeout > 0:
+        when = ceil(loop.time() + timeout)
+        return loop.call_at(when, cb)
+
+
+class TimeoutHandle:
+    """ Timeout handle """
+
+    def __init__(self, loop, timeout):
+        self._timeout = timeout
+        self._loop = loop
+        self._callbacks = []
+
+    def register(self, callback, *args, **kwargs):
+        self._callbacks.append((callback, args, kwargs))
+
+    def close(self):
+        self._callbacks.clear()
+
+    def start(self):
+        if self._timeout is not None and self._timeout > 0:
+            at = ceil(self._loop.time() + self._timeout)
+            return self._loop.call_at(at, self.__call__)
+
+    def timer(self):
+        timer = TimerContext(self._loop)
+        self.register(timer.timeout)
         return timer
 
-    def timeout(self, timeout):
-        """low resolution timeout context manager.
+    def __call__(self):
+        for cb, args, kwargs in self._callbacks:
+            try:
+                cb(*args, **kwargs)
+            except:
+                pass
 
-        timeout - value in seconds or None to disable timeout logic
-        """
-        if timeout:
-            when = self._loop_time + timeout
-            ctx = _TimeServiceTimeoutContext(when, self._loop)
-            heapq.heappush(self._scheduled, ctx)
-        else:
-            ctx = _TimeServiceTimeoutNoop()
-
-        return ctx
+        self._callbacks.clear()
 
 
-class _TimeServiceTimeoutNoop:
+class TimerNoop:
 
     def __enter__(self):
         return self
@@ -728,12 +662,11 @@ class _TimeServiceTimeoutNoop:
         return False
 
 
-class _TimeServiceTimeoutContext(TimerHandle):
+class TimerContext:
     """ Low resolution timeout context manager """
 
-    def __init__(self, when, loop):
-        super().__init__(when, self.cancel, (), loop)
-
+    def __init__(self, loop):
+        self._loop = loop
         self._tasks = []
         self._cancelled = False
 
@@ -752,20 +685,39 @@ class _TimeServiceTimeoutContext(TimerHandle):
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         if self._tasks:
-            self._tasks.pop()
+            task = self._tasks.pop()
+        else:
+            task = None
 
         if exc_type is asyncio.CancelledError and self._cancelled:
+            for task in self._tasks:
+                task.cancel()
             raise asyncio.TimeoutError from None
 
-    def cancel(self):
+        if exc_type is None and self._cancelled and task is not None:
+            task.cancel()
+
+    def timeout(self):
         if not self._cancelled:
             for task in self._tasks:
                 task.cancel()
 
-            self._tasks = []
             self._cancelled = True
 
 
+class CeilTimeout(Timeout):
+
+    def __enter__(self):
+        if self._timeout is not None:
+            self._task = asyncio.Task.current_task(loop=self._loop)
+            if self._task is None:
+                raise RuntimeError(
+                    'Timeout context manager should be used inside a task')
+            self._cancel_handler = self._loop.call_at(
+                ceil(self._loop.time() + self._timeout), self._cancel_task)
+        return self
+
+
 class HeadersMixin:
 
     _content_type = None
@@ -784,7 +736,7 @@ class HeadersMixin:
     @property
     def content_type(self, *, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
         """The value of content part for Content-Type HTTP header."""
-        raw = self.headers.get(_CONTENT_TYPE)
+        raw = self._headers.get(_CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_type
@@ -792,7 +744,7 @@ class HeadersMixin:
     @property
     def charset(self, *, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
         """The value of charset part for Content-Type HTTP header."""
-        raw = self.headers.get(_CONTENT_TYPE)
+        raw = self._headers.get(_CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_dict.get('charset')
@@ -800,7 +752,7 @@ class HeadersMixin:
     @property
     def content_length(self, *, _CONTENT_LENGTH=hdrs.CONTENT_LENGTH):
         """The value of Content-Length HTTP header."""
-        l = self.headers.get(_CONTENT_LENGTH)
+        l = self._headers.get(_CONTENT_LENGTH)
         if l is None:
             return None
         else:
