@@ -6,9 +6,10 @@ import pytest
 from multidict import CIMultiDict, MultiDict
 from yarl import URL
 
-from aiohttp.protocol import HttpVersion
+from aiohttp import HttpVersion
 from aiohttp.streams import StreamReader
 from aiohttp.test_utils import make_mocked_request
+from aiohttp.web import HTTPRequestEntityTooLarge
 
 
 @pytest.fixture
@@ -27,29 +28,26 @@ def test_ctor(make_request):
     assert 'a=1&b=2' == req.query_string
     assert CIMultiDict() == req.headers
     assert () == req.raw_headers
+    assert req.message == req._message
 
-    get = req.GET
+    get = req.query
     assert MultiDict([('a', '1'), ('b', '2')]) == get
     # second call should return the same object
-    assert get is req.GET
+    assert get is req.query
 
     assert req.keep_alive
 
     # just make sure that all lines of make_mocked_request covered
     headers = CIMultiDict(FOO='bar')
-    reader = mock.Mock()
-    writer = mock.Mock()
     payload = mock.Mock()
-    transport = mock.Mock()
+    protocol = mock.Mock()
     app = mock.Mock()
     req = make_request('GET', '/path/to?a=1&b=2', headers=headers,
-                       writer=writer, reader=reader, payload=payload,
-                       transport=transport, app=app)
+                       protocol=protocol, payload=payload, app=app)
     assert req.app is app
     assert req.content is payload
-    assert req.transport is transport
-    assert req._reader is reader
-    assert req._writer is writer
+    assert req.protocol is protocol
+    assert req.transport is protocol.transport
     assert req.headers == headers
     assert req.raw_headers == ((b'Foo', b'bar'),)
 
@@ -60,17 +58,6 @@ def test_doubleslashes(make_request):
     assert '/bar//foo/' == req.path
 
 
-def test_POST(make_request):
-    req = make_request('POST', '/')
-    with pytest.raises(RuntimeError):
-        req.POST
-
-    marker = object()
-    req._post = marker
-    assert req.POST is marker
-    assert req.POST is marker
-
-
 def test_content_type_not_specified(make_request):
     req = make_request('Get', '/')
     assert 'application/octet-stream' == req.content_type
@@ -101,7 +88,7 @@ def test_calc_content_type_on_getting_charset(make_request):
 def test_urlencoded_querystring(make_request):
     req = make_request('GET',
                        '/yandsearch?text=%D1%82%D0%B5%D0%BA%D1%81%D1%82')
-    assert {'text': 'текст'} == req.GET
+    assert {'text': 'текст'} == req.query
 
 
 def test_non_ascii_path(make_request):
@@ -111,7 +98,7 @@ def test_non_ascii_path(make_request):
 
 def test_non_ascii_raw_path(make_request):
     req = make_request('GET', '/путь')
-    assert '/%D0%BF%D1%83%D1%82%D1%8C' == req.raw_path
+    assert '/путь' == req.raw_path
 
 
 def test_content_length(make_request):
@@ -318,3 +305,45 @@ def test_cannot_clone_after_read(loop):
     yield from req.read()
     with pytest.raises(RuntimeError):
         req.clone()
+
+
+@asyncio.coroutine
+def test_make_too_big_request(loop):
+    payload = StreamReader(loop=loop)
+    large_file = 1024 ** 2 * b'x'
+    too_large_file = large_file + b'x'
+    payload.feed_data(too_large_file)
+    payload.feed_eof()
+    req = make_mocked_request('POST', '/', payload=payload)
+    with pytest.raises(HTTPRequestEntityTooLarge) as err:
+        yield from req.read()
+
+    assert err.value.status_code == 413
+
+
+@asyncio.coroutine
+def test_make_too_big_request_adjust_limit(loop):
+    payload = StreamReader(loop=loop)
+    large_file = 1024 ** 2 * b'x'
+    too_large_file = large_file + b'x'
+    payload.feed_data(too_large_file)
+    payload.feed_eof()
+    max_size = 1024**2 + 2
+    req = make_mocked_request('POST', '/', payload=payload,
+                              client_max_size=max_size)
+    txt = yield from req.read()
+    assert len(txt) == 1024**2 + 1
+
+
+@asyncio.coroutine
+def test_make_too_big_request_limit_None(loop):
+    payload = StreamReader(loop=loop)
+    large_file = 1024 ** 2 * b'x'
+    too_large_file = large_file + b'x'
+    payload.feed_data(too_large_file)
+    payload.feed_eof()
+    max_size = None
+    req = make_mocked_request('POST', '/', payload=payload,
+                              client_max_size=max_size)
+    txt = yield from req.read()
+    assert len(txt) == 1024**2 + 1
